"""Comprehensive test suite for Template Mode functionality.

This suite focuses on the scenarios highlighted during the review of the
`feat/t-parse-v1` branch while keeping the behaviour of this branch's
implementation intact. The coverage mirrors that suite's structure:

1. Core behaviours around static templates, PYOBJ handling, and tree splicing
2. Error handling and configuration edge cases
3. A graphics DSL demonstrating mixed static content, typed placeholders, and
   tree splicing in more realistic programs
"""

import unittest

from t_lark import Lark, Tree
from t_lark.exceptions import ConfigurationError, UnexpectedToken


class TestTemplateModeBasics(unittest.TestCase):
    """Basic template mode functionality tests."""

    def test_static_only(self):
        """Static template should parse identically to the plain string."""
        grammar = r"""
        start: NUMBER
        %import common.NUMBER
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        static = t"42"
        self.assertEqual(parser.parse(static), parser.parse("42"))

    def test_pyobj_untyped(self):
        """PYOBJ should accept arbitrary Python objects without a type map."""
        grammar = r"""
        %import template (PYOBJ)
        start: "value:" PYOBJ
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        self.assertIsNotNone(parser.parse(t"value:{42}"))
        self.assertIsNotNone(parser.parse(t"value:{'hello'}"))
        self.assertIsNotNone(parser.parse(t"value:{[1, 2, 3]}"))

    def test_pyobj_typed_success(self):
        """Typed PYOBJ placeholders should accept objects of the expected type."""
        grammar = r"""
        %import template (PYOBJ)
        start: PYOBJ[num]
        """
        parser = Lark(grammar, parser="earley", lexer="template", pyobj_types={'num': int})

        tree = parser.parse(t"{7}")
        self.assertIsInstance(tree, Tree)

    def test_tree_splicing(self):
        """Pre-built trees should splice seamlessly into larger parses."""
        grammar = r"""
        ?start: expr
        ?expr: term
             | expr "+" term -> add
        ?term: NUMBER
        %import common.NUMBER
        """
        parser = Lark(grammar, parser="earley", lexer="template", start=['start', 'expr'])

        fragment = parser.parse("1+2", start='expr')
        self.assertEqual(fragment.data, 'add')

        combined = parser.parse(t"{fragment}", start='start')
        self.assertEqual(combined.data, 'add')

    def test_consecutive_objects(self):
        """Adjacent placeholders should be handled without extra separators."""
        grammar = r"""
        %import template (PYOBJ)
        start: PYOBJ PYOBJ
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        tree = parser.parse(t"{1}{2}")
        self.assertIsInstance(tree, Tree)

    def test_mixed_static_content(self):
        """Mixing static text, tree splicing, and PYOBJ interpolation should work."""
        grammar = r"""
        %import template (PYOBJ)
        start: "static" PYOBJ expr
        expr: NUMBER
        %import common.NUMBER
        %ignore " "
        """
        parser = Lark(grammar, parser="earley", lexer="template", start=['start', 'expr'])

        chunk = parser.parse("42", start='expr')
        result = parser.parse(t"static {100} {chunk}", start='start')
        self.assertEqual(result.data, 'start')

    def test_source_info_absent(self):
        """Template objects without source metadata should still parse."""
        grammar = r"""
        %import template (PYOBJ)
        start: PYOBJ
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        self.assertIsNotNone(parser.parse(t"{42}"))


class TestTemplateModeErrors(unittest.TestCase):
    """Error handling and configuration coverage."""

    def test_pyobj_typed_type_error(self):
        """Typed placeholders should raise informative errors for mismatched types."""
        grammar = r"""
        %import template (PYOBJ)
        start: PYOBJ[num]
        """
        parser = Lark(grammar, parser="earley", lexer="template", pyobj_types={'num': int})

        with self.assertRaises(TypeError) as ctx:
            parser.parse(t"{'not an int'}")

        message = str(ctx.exception)
        self.assertIn("PYOBJ[num]", message)
        self.assertIn("str", message)

    def test_missing_pyobj_types_configuration(self):
        """Using typed PYOBJ without providing a mapping should fail fast."""
        grammar = r"""
        %import template (PYOBJ)
        start: PYOBJ[image]
        """
        with self.assertRaises(ConfigurationError):
            Lark(grammar, parser="earley", lexer="template")

    def test_error_no_pyobj(self):
        """Interpolating an object where none are allowed should raise."""
        grammar = r"""
        start: NUMBER
        %import common.NUMBER
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        with self.assertRaises(UnexpectedToken):
            parser.parse(t"{5}")

    def test_error_wrong_tree_label(self):
        """Splicing a tree the grammar cannot produce should raise."""
        grammar = r"""
        start: NUMBER
        %import common.NUMBER
        """
        parser = Lark(grammar, parser="earley", lexer="template")

        wrong = Tree('unknown', [])
        with self.assertRaises((UnexpectedToken, ValueError)) as ctx:
            parser.parse(t"{wrong}")

        self.assertIn("unknown", str(ctx.exception))

    def test_requires_earley_parser(self):
        """Template mode requires the Earley parser backend."""
        grammar = 'start: "x"'
        with self.assertRaises(ConfigurationError):
            Lark(grammar, parser="lalr", lexer="template")


class TestPaintDSL(unittest.TestCase):
    """Realistic DSL exercising typed placeholders, splicing, and errors."""

    class Image:
        def __init__(self, path: str) -> None:
            self.path = path

        def __repr__(self) -> str:  # pragma: no cover - debug helper
            return f"Image({self.path!r})"

        def __eq__(self, other) -> bool:
            return isinstance(other, type(self)) and self.path == other.path

    GRAMMAR = r"""
    %import template (PYOBJ)

    start: object+

    object: "object" "{" "stroke:" paint "fill:" paint "}"

    paint: color
         | image

    color: NUMBER "," NUMBER "," NUMBER -> color

    image: PYOBJ[image] -> image

    %import common.NUMBER
    %ignore " "
    """

    def setUp(self):
        self.parser = Lark(
            self.GRAMMAR,
            parser="earley",
            lexer="template",
            pyobj_types={'image': self.Image},
            start=['start', 'color'],
        )

    def test_static_only_parse(self):
        """Purely static template should parse into one object with colors."""
        program = t"object {{ stroke: 255,0,0 fill: 0,255,0 }}"
        tree = self.parser.parse(program, start='start')

        self.assertEqual(tree.data, 'start')
        self.assertEqual(len(tree.children), 1)

        obj = tree.children[0]
        self.assertEqual(obj.data, 'object')

        stroke_paint = obj.children[0]
        fill_paint = obj.children[1]
        self.assertEqual(stroke_paint.children[0].data, 'color')
        self.assertEqual(fill_paint.children[0].data, 'color')

    def test_interpolate_image_objects(self):
        """Interpolated Image objects should surface as typed tokens."""
        texture = self.Image("wood_texture.png")
        gradient = self.Image("gradient.png")

        program = t"object {{ stroke: {texture} fill: {gradient} }}"
        tree = self.parser.parse(program, start='start')

        obj = tree.children[0]
        stroke_image = obj.children[0].children[0]
        fill_image = obj.children[1].children[0]

        self.assertEqual(stroke_image.data, 'image')
        self.assertEqual(fill_image.data, 'image')

        stroke_token = stroke_image.children[0]
        fill_token = fill_image.children[0]

        self.assertEqual(stroke_token.type, 'PYOBJ__IMAGE')
        self.assertEqual(fill_token.type, 'PYOBJ__IMAGE')
        self.assertIs(stroke_token.value, texture)
        self.assertIs(fill_token.value, gradient)

    def test_type_mismatch_error(self):
        """Passing an incompatible type should raise TypeError with context."""
        program = t"object {{ stroke: {'not an image'} fill: 0,0,255 }}"

        with self.assertRaises(TypeError) as ctx:
            self.parser.parse(program, start='start')

        message = str(ctx.exception)
        self.assertIn("PYOBJ[image]", message)
        self.assertIn("str", message)

    def test_type_mismatch_integer(self):
        """Type mismatch should report the offending integer type."""
        program = t"object {{ stroke: {42} fill: 0,0,0 }}"

        with self.assertRaises(TypeError) as ctx:
            self.parser.parse(program, start='start')

        message = str(ctx.exception)
        self.assertIn("PYOBJ[image]", message)
        self.assertIn("int", message)

    def test_tree_splicing_color(self):
        """Colors parsed separately can be spliced into paint positions."""
        red = self.parser.parse("255,0,0", start='color')
        program = t"object {{ stroke: {red} fill: {red} }}"

        tree = self.parser.parse(program, start='start')
        obj = tree.children[0]

        stroke_color = obj.children[0].children[0]
        fill_color = obj.children[1].children[0]
        self.assertEqual(stroke_color.data, 'color')
        self.assertEqual(fill_color.data, 'color')

    def test_tree_splicing_with_control_flow(self):
        """Control flow assembling templates should still produce valid trees."""
        red = self.parser.parse("255,0,0", start='color')
        blue = self.parser.parse("0,0,255", start='color')
        green = self.parser.parse("0,255,0", start='color')

        colors = [red, green, blue]
        palette = []

        for index, color in enumerate(colors):
            if index % 2 == 0:
                palette.append(t"object {{ stroke: {color} fill: {color} }}")
            else:
                palette.append(t"object {{ stroke: {color} fill: 0,0,0 }}")

        program = palette[0]
        for fragment in palette[1:]:
            program = program + fragment
        tree = self.parser.parse(program, start='start')

        self.assertEqual(len(tree.children), 3)
        first = tree.children[0]
        second = tree.children[1]

        self.assertEqual(first.children[0].children[0].data, 'color')
        self.assertEqual(first.children[1].children[0].data, 'color')
        self.assertEqual(second.children[0].children[0].data, 'color')
        self.assertEqual(second.children[1].children[0].data, 'color')

    def test_mixed_static_object_tree(self):
        """Mix static segments, spliced trees, and interpolated images."""
        red = self.parser.parse("255,0,0", start='color')
        texture = self.Image("pattern.png")

        program = (
            t"object {{ stroke: 128,128,128 fill: 64,64,64 }} "
            + t"object {{ stroke: {red} fill: {texture} }} "
            + t"object {{ stroke: {texture} fill: 0,255,128 }}"
        )

        tree = self.parser.parse(program, start='start')
        self.assertEqual(len(tree.children), 3)

        obj1, obj2, obj3 = tree.children
        self.assertEqual(obj1.children[0].children[0].data, 'color')
        self.assertEqual(obj1.children[1].children[0].data, 'color')
        self.assertEqual(obj2.children[0].children[0].data, 'color')
        self.assertEqual(obj2.children[1].children[0].data, 'image')
        self.assertEqual(obj3.children[0].children[0].data, 'image')
        self.assertEqual(obj3.children[1].children[0].data, 'color')

    def test_error_wrong_tree_label(self):
        """Splicing an unsupported tree should raise with the label mentioned."""
        wrong_tree = Tree('circle', [])
        program = t"object {{ stroke: {wrong_tree} fill: 0,0,0 }}"

        with self.assertRaises(ValueError) as ctx:
            self.parser.parse(program, start='start')

        self.assertIn("circle", str(ctx.exception))

    def test_error_malformed_static_syntax(self):
        """Static syntax errors should be propagated as UnexpectedToken."""
        program = t"object {{ stroke: 255,0 fill: 0,0,0 }}"

        with self.assertRaises(UnexpectedToken):
            self.parser.parse(program, start='start')

    def test_multiple_objects_mixed_paints(self):
        """Larger programs mixing colors and images should parse cleanly."""
        red = self.parser.parse("255,0,0", start='color')
        img1 = self.Image("texture1.png")
        img2 = self.Image("texture2.png")

        program = (
            t"object {{ stroke: 255,255,255 fill: 0,0,0 }} "
            + t"object {{ stroke: {red} fill: 100,100,100 }} "
            + t"object {{ stroke: {img1} fill: {img2} }} "
            + t"object {{ stroke: 50,50,50 fill: {red} }} "
            + t"object {{ stroke: {img1} fill: 200,200,200 }}"
        )

        tree = self.parser.parse(program, start='start')
        self.assertEqual(len(tree.children), 5)

        for obj in tree.children:
            self.assertEqual(obj.data, 'object')
            self.assertEqual(len(obj.children), 2)
            for paint in obj.children:
                kind = paint.children[0].data
                self.assertIn(kind, {'color', 'image'})

    def test_image_object_preservation(self):
        """Image objects should be preserved verbatim through the parse."""
        image = self.Image("test.png")
        program = t"object {{ stroke: {image} fill: 0,0,0 }}"

        tree = self.parser.parse(program, start='start')
        obj = tree.children[0]
        stroke_image = obj.children[0].children[0]
        image_token = stroke_image.children[0]

        self.assertIs(image_token.value, image)
        self.assertEqual(image_token.value.path, "test.png")

    def test_static_equivalent_to_plain_string(self):
        """Static-only template should match parsing the plain string."""
        tmpl_tree = self.parser.parse(t"object {{ stroke: 255,0,0 fill: 0,255,0 }}", start='start')
        plain_tree = self.parser.parse("object { stroke: 255,0,0 fill: 0,255,0 }", start='start')

        self.assertEqual(tmpl_tree.data, plain_tree.data)
        self.assertEqual(len(tmpl_tree.children), len(plain_tree.children))


if __name__ == '__main__':
    unittest.main()



"""Utilities for Lark's template parsing mode."""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, Mapping, Optional, Sequence, Set, Tuple, cast

from .common import LexerConf, ParserConf
from .grammar import NonTerminal, Rule, RuleOptions, Terminal
from .lexer import BasicLexer, LexerThread, PatternPlaceholder, PatternTree, TerminalDef, Token
from .tree import Tree
from .utils import TextSlice


_SANITIZE_RE = re.compile(r"[^0-9A-Z_]")


def basic_lexer_for_static(lexer_conf: LexerConf) -> BasicLexer:
    """Create a BasicLexer configured to ignore template-only terminals."""

    static_terminals = [
        term for term in lexer_conf.terminals
        if not isinstance(term.pattern, (PatternTree, PatternPlaceholder))
    ]

    filtered_conf = LexerConf(
        static_terminals,
        lexer_conf.re_module,
        lexer_conf.ignore,
        lexer_conf.postlex,
        lexer_conf.callbacks,
        lexer_conf.g_regex_flags,
        skip_validation=True,
        use_bytes=lexer_conf.use_bytes,
        strict=lexer_conf.strict,
    )
    filtered_conf.terminals_by_name = {t.name: t for t in static_terminals}
    filtered_conf.lexer_type = lexer_conf.lexer_type
    return BasicLexer(filtered_conf)


def _tree_injection_callback(children):
    token ,= children
    return token.value


@dataclass
class SourceInfo:
    """Represents source metadata for template segments/interpolations."""

    filename: str
    text: str
    segment_spans: Sequence[Tuple[int, int]]
    interpolation_spans: Sequence[Tuple[int, int]]


@dataclass
class TemplateContext:
    """Context required to tokenize a template."""

    lexer_conf: LexerConf
    tree_terminal_map: Mapping[str, str]
    typed_terminals: Mapping[str, str]
    pyobj_types: Mapping[str, type | tuple[type, ...]]
    source_info: Optional[SourceInfo] = None


class _TemplateLexer:
    """Adapter that exposes `lex` for Earley parser consumption."""

    def __init__(self, template, ctx: TemplateContext):
        self._template = template
        self._ctx = ctx
        self._used = False

    def lex(self, parser_state):
        if self._used:
            return iter(())
        self._used = True
        token_iter: Iterable[Token] = _iterate_template_tokens(self._template, self._ctx, parser_state)
        postlex = self._ctx.lexer_conf.postlex
        if postlex is not None:
            token_iter = postlex.process(iter(token_iter))
        return iter(token_iter)


def tokenize_template(template, ctx: TemplateContext) -> _TemplateLexer:
    """Create a lexer adapter for a template object."""

    return _TemplateLexer(template, ctx)


def _iterate_template_tokens(template, ctx: TemplateContext, parser_state) -> Iterator[Token]:
    basic_lexer = basic_lexer_for_static(ctx.lexer_conf)
    strings = template.strings
    interpolations = template.interpolations

    for index, static_str in enumerate(strings):
        if static_str:
            text_slice = _segment_text_slice(static_str, index, ctx)
            thread = LexerThread.from_text(basic_lexer, text_slice)
            yield from thread.lex(parser_state)

        if index < len(interpolations):
            interp = interpolations[index]
            yield _build_interpolation_token(interp.value, ctx, index)


def _segment_text_slice(static_str: str, index: int, ctx: TemplateContext) -> TextSlice[str]:
    info = ctx.source_info
    if info is None:
        return TextSlice(static_str, 0, len(static_str))

    start, end = info.segment_spans[index]
    return TextSlice(info.text, start, end)


def _build_interpolation_token(value, ctx: TemplateContext, index: int) -> Token:
    info = ctx.source_info
    if info is None:
        meta: Dict[str, Optional[int]] = {
            "start_pos": None,
            "end_pos": None,
            "line": None,
            "column": None,
            "end_line": None,
            "end_column": None,
        }
    else:
        start, end = info.interpolation_spans[index]
        meta = cast(Dict[str, Optional[int]], _offset_to_meta(info.text, start, end))

    if isinstance(value, Tree):
        term_name = ctx.tree_terminal_map.get(value.data)
        if term_name is None:
            raise ValueError(f"Cannot splice Tree('{value.data}'): grammar does not produce this label")

        token = _tree_token(term_name, value, meta)
        return token

    terminal_name, _ = _resolve_pyobj_terminal(value, ctx)
    token_type = terminal_name or "PYOBJ"
    return Token(token_type, value, **meta)


def _tree_token(term_name: str, tree_value: Tree, meta: Dict[str, Optional[int]]) -> Token:
    if hasattr(tree_value, "meta") and tree_value.meta:
        m = tree_value.meta
        return Token(
            term_name,
            tree_value,
            start_pos=getattr(m, "start_pos", meta["start_pos"]),
            line=getattr(m, "line", meta["line"]),
            column=getattr(m, "column", meta["column"]),
            end_pos=getattr(m, "end_pos", meta["end_pos"]),
            end_line=getattr(m, "end_line", meta["end_line"]),
            end_column=getattr(m, "end_column", meta["end_column"]),
        )

    return Token(term_name, tree_value, **meta)


def _resolve_pyobj_terminal(value, ctx: TemplateContext) -> Tuple[Optional[str], Optional[str]]:
    matches: list[Tuple[int, str, str]] = []
    for type_name, expected_type in ctx.pyobj_types.items():
        terminal = ctx.typed_terminals.get(type_name)
        if terminal is None:
            continue

        candidates: Sequence[type]
        if isinstance(expected_type, tuple):
            candidates = expected_type
        else:
            candidates = (expected_type,)

        for candidate in candidates:
            if isinstance(value, candidate):
                distance = _inheritance_distance(candidate, type(value))
                matches.append((distance, type_name, terminal))
                break

    if not matches:
        return None, None

    matches.sort(key=lambda item: (item[0], item[1]))
    _, type_name, terminal_name = matches[0]
    return terminal_name, type_name


def _inheritance_distance(expected: type, actual: type) -> int:
    try:
        return actual.mro().index(expected)
    except ValueError:
        return len(actual.mro())


def _offset_to_meta(text: str, start: int, end: int) -> Dict[str, int]:
    line = text.count("\n", 0, start) + 1
    line_start = text.rfind("\n", 0, start) + 1
    column = start - line_start + 1

    lines_in_span = text.count("\n", start, end)
    end_line = line + lines_in_span

    if lines_in_span == 0:
        end_column = column + (end - start)
    else:
        last_line_start = text.rfind("\n", 0, end) + 1
        end_column = end - last_line_start + 1

    return {
        "start_pos": start,
        "end_pos": end,
        "line": line,
        "column": column,
        "end_line": end_line,
        "end_column": end_column,
    }


def splice_inserted_trees(node):
    """Replace TREE__ tokens with the Tree objects produced during parsing."""

    if not isinstance(node, Tree):
        return

    new_children = []
    for child in node.children:
        if isinstance(child, Token) and child.type.startswith("TREE__") and isinstance(child.value, Tree):
            splice_inserted_trees(child.value)
            new_children.append(child.value)
        else:
            if isinstance(child, Tree):
                splice_inserted_trees(child)
            new_children.append(child)

    node.children = new_children


def _make_tree_terminal_name(label: str) -> str:
    upper = label.upper()
    safe = _SANITIZE_RE.sub('_', upper)
    return f"TREE__{safe}"


def augment_grammar_for_template_mode(lexer_conf: LexerConf, parser_conf: ParserConf) -> Mapping[str, str]:
    """Augment grammar with tree injection rules for template parsing."""

    existing = getattr(parser_conf, "template_tree_terminals", None)
    if existing is not None:
        return existing

    label_map: Dict[str, str] = {}
    labels_per_nonterminal: Dict[NonTerminal, Set[str]] = {}

    for rule in parser_conf.rules:
        labels = labels_per_nonterminal.setdefault(rule.origin, set())
        label = rule.alias if rule.alias else rule.origin.name
        labels.add(str(label))

    terminals = list(lexer_conf.terminals)
    terminals_by_name = dict(lexer_conf.terminals_by_name)

    for labels in labels_per_nonterminal.values():
        for label in labels:
            term_name = _make_tree_terminal_name(label)
            label_map[label] = term_name
            if term_name in terminals_by_name:
                continue
            terminals.append(TerminalDef(term_name, PatternTree(label)))
            terminals_by_name[term_name] = terminals[-1]

    lexer_conf.terminals = terminals
    lexer_conf.terminals_by_name = terminals_by_name

    existing_expansions: Set[Tuple[NonTerminal, Tuple[str, ...]]] = set()
    max_order: Dict[NonTerminal, int] = {}
    for rule in parser_conf.rules:
        expansion = tuple(symbol.name for symbol in rule.expansion)
        existing_expansions.add((rule.origin, expansion))
        max_order[rule.origin] = max(max_order.get(rule.origin, -1), rule.order)

    new_rules: list[Rule] = []

    for origin, labels in labels_per_nonterminal.items():
        order = max_order.get(origin, -1) + 1
        for label in labels:
            term_name = label_map[label]
            key = (origin, (term_name,))
            if key in existing_expansions:
                continue
            symbol = Terminal(term_name)
            new_rules.append(Rule(origin, [symbol], order, alias=None, options=RuleOptions(expand1=True)))
            existing_expansions.add(key)
            order += 1

    if new_rules:
        parser_conf.rules.extend(new_rules)
        for rule in new_rules:
            parser_conf.callbacks[rule] = _tree_injection_callback

    parser_conf.template_tree_terminals = label_map
    return label_map



from functools import lru_cache
from importlib import import_module
from typing import Any, Callable, Dict, Optional, Collection, Union, TYPE_CHECKING

from .exceptions import ConfigurationError, GrammarError, UnexpectedInput, UnexpectedToken, assert_config
from .utils import get_regexp_width, Serialize, TextOrSlice, TextSlice
from .lexer import LexerThread, BasicLexer, ContextualLexer, Lexer, Token, PatternPlaceholder
from .parsers import earley, xearley, cyk
from .parsers.lalr_parser import LALR_Parser
from .tree import Tree
from .common import LexerConf, ParserConf, _ParserArgType, _LexerArgType
from .template_mode import (
    TemplateContext,
    augment_grammar_for_template_mode,
    tokenize_template,
    splice_inserted_trees,
    basic_lexer_for_static,
)

if TYPE_CHECKING:
    from .parsers.lalr_analysis import ParseTableBase


###{standalone

from functools import lru_cache
from importlib import import_module

def _wrap_lexer(lexer_class):
    future_interface = getattr(lexer_class, '__future_interface__', 0)
    if future_interface == 2:
        return lexer_class
    elif future_interface == 1:
        class CustomLexerWrapper1(Lexer):
            def __init__(self, lexer_conf):
                self.lexer = lexer_class(lexer_conf)
            def lex(self, lexer_state, parser_state):
                if not lexer_state.text.is_complete_text():
                    raise TypeError("Interface=1 Custom Lexer don't support TextSlice")
                lexer_state.text = lexer_state.text
                return self.lexer.lex(lexer_state, parser_state)
        return CustomLexerWrapper1
    elif future_interface == 0:
        class CustomLexerWrapper0(Lexer):
            def __init__(self, lexer_conf):
                self.lexer = lexer_class(lexer_conf)

            def lex(self, lexer_state, parser_state):
                if not lexer_state.text.is_complete_text():
                    raise TypeError("Interface=0 Custom Lexer don't support TextSlice")
                return self.lexer.lex(lexer_state.text.text)
        return CustomLexerWrapper0
    else:
        raise ValueError(f"Unknown __future_interface__ value {future_interface}, integer 0-2 expected")


def _deserialize_parsing_frontend(data, memo, lexer_conf, callbacks, options):
    parser_conf = ParserConf.deserialize(data['parser_conf'], memo)
    cls = (options and options._plugins.get('LALR_Parser')) or LALR_Parser
    parser = cls.deserialize(data['parser'], memo, callbacks, options.debug)
    parser_conf.callbacks = callbacks
    return ParsingFrontend(lexer_conf, parser_conf, options, parser=parser)


_parser_creators: 'Dict[str, Callable[[LexerConf, Any, Any], Any]]' = {}

@lru_cache(maxsize=1)
def _string_template_type() -> Optional[type[Any]]:
    try:
        module = import_module("string.templatelib")
    except ModuleNotFoundError:
        return None
    template_cls = getattr(module, "Template")
    if not isinstance(template_cls, type):
        raise TypeError("string.templatelib.Template is expected to be a type")
    return template_cls


class ParsingFrontend(Serialize):
    __serialize_fields__ = 'lexer_conf', 'parser_conf', 'parser'

    lexer_conf: LexerConf
    parser_conf: ParserConf
    options: Any

    def __init__(self, lexer_conf: LexerConf, parser_conf: ParserConf, options, parser=None):
        self.parser_conf = parser_conf
        self.lexer_conf = lexer_conf
        self.options = options

        # Set-up parser
        if parser:  # From cache
            self.parser = parser
        else:
            create_parser = _parser_creators.get(parser_conf.parser_type)
            assert create_parser is not None, "{} is not supported in standalone mode".format(
                    parser_conf.parser_type
                )
            self.parser = create_parser(lexer_conf, parser_conf, options)

        # Set-up lexer
        lexer_type = lexer_conf.lexer_type
        self.skip_lexer = False
        if lexer_type in ('dynamic', 'dynamic_complete'):
            assert lexer_conf.postlex is None
            self.skip_lexer = True
            return

        if isinstance(lexer_type, type):
            assert issubclass(lexer_type, Lexer)
            self.lexer = _wrap_lexer(lexer_type)(lexer_conf)
        elif isinstance(lexer_type, str):
            create_lexer = {
                'basic': create_basic_lexer,
                'contextual': create_contextual_lexer,
            }[lexer_type]
            self.lexer = create_lexer(lexer_conf, self.parser, lexer_conf.postlex, options)
        else:
            raise TypeError("Bad value for lexer_type: {lexer_type}")

        if lexer_conf.postlex:
            self.lexer = PostLexConnector(self.lexer, lexer_conf.postlex)

    def _verify_start(self, start=None):
        if start is None:
            start_decls = self.parser_conf.start
            if len(start_decls) > 1:
                raise ConfigurationError("Lark initialized with more than 1 possible start rule. Must specify which start rule to parse", start_decls)
            start ,= start_decls
        elif start not in self.parser_conf.start:
            raise ConfigurationError("Unknown start rule %s. Must be one of %r" % (start, self.parser_conf.start))
        return start

    def _make_lexer_thread(self, text: Optional[TextOrSlice]) -> Union[TextOrSlice, LexerThread, None]:
        cls = (self.options and self.options._plugins.get('LexerThread')) or LexerThread
        return text if self.skip_lexer else cls(self.lexer, None) if text is None else cls.from_text(self.lexer, text)

    def parse(self, text: Optional[TextOrSlice], start=None, on_error=None):
        if self.lexer_conf.lexer_type in ("dynamic", "dynamic_complete"):
            if isinstance(text, TextSlice) and not text.is_complete_text():
                raise TypeError(f"Lexer {self.lexer_conf.lexer_type} does not support text slices.")

        chosen_start = self._verify_start(start)
        kw = {} if on_error is None else {'on_error': on_error}
        stream = self._make_lexer_thread(text)
        return self.parser.parse(stream, chosen_start, **kw)

    def parse_interactive(self, text: Optional[TextOrSlice]=None, start=None):
        # TODO BREAK - Change text from Optional[str] to text: str = ''.
        #   Would break behavior of exhaust_lexer(), which currently raises TypeError, and after the change would just return []
        chosen_start = self._verify_start(start)
        if self.parser_conf.parser_type != 'lalr':
            raise ConfigurationError("parse_interactive() currently only works with parser='lalr' ")
        stream = self._make_lexer_thread(text)
        return self.parser.parse_interactive(stream, chosen_start)


class _StringLexer:
    """Minimal lexer wrapper for template frontend when parsing plain strings."""

    def __init__(self, lexer_conf: LexerConf, text: TextSlice) -> None:
        self._lexer = basic_lexer_for_static(lexer_conf)
        self._thread = LexerThread.from_text(self._lexer, text)
        self._postlex = lexer_conf.postlex
        self._used = False

    def lex(self, parser_state):
        if self._used:
            return iter(())
        self._used = True
        tokens = self._thread.lex(parser_state)
        if self._postlex is not None:
            tokens = self._postlex.process(tokens)
        return tokens


class TemplateEarleyFrontend:
    """Parsing frontend for Python template literals using the Earley parser."""

    def __init__(self, lexer_conf: LexerConf, parser_conf: ParserConf, options) -> None:
        self.lexer_conf = lexer_conf
        self.parser_conf = parser_conf
        self.options = options

        self.tree_terminal_map = dict(augment_grammar_for_template_mode(lexer_conf, parser_conf))
        self._tree_terminal_inverse = {terminal: label for label, terminal in self.tree_terminal_map.items()}

        provided_types = getattr(options, 'pyobj_types', {}) or {}
        self.pyobj_types: Dict[str, type | tuple[type, ...]] = dict(provided_types)

        self.typed_terminals_by_type: Dict[str, str] = {}
        self.typed_terminals_by_name: Dict[str, str] = {}
        self._typed_expected_types: Dict[str, type | tuple[type, ...]] = {}

        for termdef in lexer_conf.terminals:
            pattern = getattr(termdef, 'pattern', None)
            if isinstance(pattern, PatternPlaceholder) and pattern.expected_type:
                type_name = pattern.expected_type
                self.typed_terminals_by_type[type_name] = termdef.name
                self.typed_terminals_by_name[termdef.name] = type_name

        missing = sorted(name for name in self.typed_terminals_by_type if name not in self.pyobj_types)
        if missing:
            raise ConfigurationError("Missing pyobj_types entries for typed placeholders: %s" % ', '.join(missing))

        for term_name, type_name in self.typed_terminals_by_name.items():
            self._typed_expected_types[term_name] = self.pyobj_types[type_name]

        term_matcher = self._create_term_matcher()

        from .parsers.earley import Parser as EarleyParser

        resolve_ambiguity = options.ambiguity == 'resolve'
        debug = options.debug if options else False
        tree_class = options.tree_class or Tree if options.ambiguity != 'forest' else None

        self.parser = EarleyParser(
            lexer_conf,
            parser_conf,
            term_matcher,
            resolve_ambiguity=resolve_ambiguity,
            debug=debug,
            tree_class=tree_class,
            ordered_sets=getattr(options, 'ordered_sets', True),
        )

    def _create_term_matcher(self):
        typed_by_name = self.typed_terminals_by_name
        typed_expected = self._typed_expected_types

        def term_match(term, token):
            if not isinstance(token, Token):
                return False

            term_name = term.name

            if term_name == 'PYOBJ':
                if token.type == 'PYOBJ':
                    return True
                if token.type in typed_by_name:
                    token.type = 'PYOBJ'
                    return True
                return False

            if term_name.startswith('PYOBJ__'):
                expected = typed_expected.get(term_name)
                type_name = typed_by_name.get(term_name)

                if type_name is None:
                    return False

                if token.type not in ('PYOBJ', term_name) and token.type not in typed_by_name:
                    return False

                self._check_pyobj_type(expected, type_name, token.value)
                token.type = term_name
                return True

            if term_name.startswith('TREE__'):
                if token.type != term_name:
                    return False
                if not isinstance(token.value, Tree):
                    return False
                expected_label = self._tree_terminal_inverse.get(term_name)
                return expected_label is None or token.value.data == expected_label

            return token.type == term_name

        return term_match

    @staticmethod
    def _format_expected_type(expected: type | tuple[type, ...]) -> str:
        if isinstance(expected, tuple):
            return ' or '.join(t.__name__ for t in expected)
        return expected.__name__

    def _check_pyobj_type(self, expected: Optional[type | tuple[type, ...]], type_name: str, value: Any) -> None:
        if expected is None:
            return
        if isinstance(value, expected):
            return
        expected_repr = self._format_expected_type(expected)
        raise TypeError(f"Expected {expected_repr} for PYOBJ[{type_name}], got {type(value).__name__}")

    def _verify_start(self, start=None):
        if start is None:
            start_candidates = self.parser_conf.start
            if len(start_candidates) > 1:
                raise ConfigurationError(
                    "Lark initialized with more than 1 possible start rule. Must specify which start rule to parse",
                    start_candidates,
                )
            start ,= start_candidates
        elif start not in self.parser_conf.start:
            raise ConfigurationError(f"Unknown start rule {start}. Must be one of {self.parser_conf.start}")
        return start

    def parse(self, input_data, start=None, on_error=None):
        chosen_start = self._verify_start(start)

        template_cls = _string_template_type()

        if template_cls is not None and isinstance(input_data, template_cls):
            lexer = self._tokenize_template(input_data)
        else:
            lexer = self._lex_string(input_data)

        try:
            tree = self.parser.parse(lexer, chosen_start)
        except UnexpectedInput as exc:
            self._enhance_error(exc, input_data)
            raise

        if tree is not None:
            splice_inserted_trees(tree)

        return tree

    def _tokenize_template(self, template: Any):
        ctx = TemplateContext(
            lexer_conf=self.lexer_conf,
            tree_terminal_map=self.tree_terminal_map,
            typed_terminals=self.typed_terminals_by_type,
            pyobj_types=self.pyobj_types,
            source_info=getattr(template, 'source_info', None),
        )
        return tokenize_template(template, ctx)

    def _lex_string(self, text):
        text_slice = TextSlice.cast_from(text)
        return _StringLexer(self.lexer_conf, text_slice)

    def _enhance_error(self, exc: UnexpectedInput, input_data) -> None:
        template_cls = _string_template_type()

        if template_cls is None or not isinstance(input_data, template_cls):
            return

        if not isinstance(exc, UnexpectedToken):
            return

        token = exc.token
        if token is None:
            return

        line = getattr(exc, 'line', '?')
        column = getattr(exc, 'column', '?')
        original = exc.args[0] if exc.args else ''

        token_type = token.type
        if token_type.startswith('PYOBJ'):
            exc.args = (
                f"Interpolated Python object at {line}:{column} not allowed here (no PYOBJ placeholder in this context). "
                f"Original: {original}",
            )
        elif token_type.startswith('TREE__'):
            label = self._tree_terminal_inverse.get(token_type, token_type[len('TREE__'):].lower())
            exc.args = (
                f"Interpolated Tree('{label}') at {line}:{column} not valid in this context. Original: {original}",
            )
def _validate_frontend_args(parser, lexer) -> None:
    assert_config(parser, ('lalr', 'earley', 'cyk'))
    if not isinstance(lexer, type):     # not custom lexer?
        expected = {
            'lalr': ('basic', 'contextual'),
            'earley': ('basic', 'dynamic', 'dynamic_complete', 'template'),
            'cyk': ('basic', ),
         }[parser]
        assert_config(lexer, expected, 'Parser %r does not support lexer %%r, expected one of %%s' % parser)


def _get_lexer_callbacks(transformer, terminals):
    result = {}
    for terminal in terminals:
        callback = getattr(transformer, terminal.name, None)
        if callback is not None:
            result[terminal.name] = callback
    return result

class PostLexConnector:
    def __init__(self, lexer, postlexer):
        self.lexer = lexer
        self.postlexer = postlexer

    def lex(self, lexer_state, parser_state):
        i = self.lexer.lex(lexer_state, parser_state)
        return self.postlexer.process(i)



def create_basic_lexer(lexer_conf, parser, postlex, options) -> BasicLexer:
    cls = (options and options._plugins.get('BasicLexer')) or BasicLexer
    return cls(lexer_conf)

def create_contextual_lexer(lexer_conf: LexerConf, parser, postlex, options) -> ContextualLexer:
    cls = (options and options._plugins.get('ContextualLexer')) or ContextualLexer
    parse_table: ParseTableBase[int] = parser._parse_table
    states: Dict[int, Collection[str]] = {idx:list(t.keys()) for idx, t in parse_table.states.items()}
    always_accept: Collection[str] = postlex.always_accept if postlex else ()
    return cls(lexer_conf, states, always_accept=always_accept)

def create_lalr_parser(lexer_conf: LexerConf, parser_conf: ParserConf, options=None) -> LALR_Parser:
    debug = options.debug if options else False
    strict = options.strict if options else False
    cls = (options and options._plugins.get('LALR_Parser')) or LALR_Parser
    return cls(parser_conf, debug=debug, strict=strict)

_parser_creators['lalr'] = create_lalr_parser

###}

class EarleyRegexpMatcher:
    def __init__(self, lexer_conf):
        self.regexps = {}
        for t in lexer_conf.terminals:
            regexp = t.pattern.to_regexp()
            try:
                width = get_regexp_width(regexp)[0]
            except ValueError:
                raise GrammarError("Bad regexp in token %s: %s" % (t.name, regexp))
            else:
                if width == 0:
                    raise GrammarError("Dynamic Earley doesn't allow zero-width regexps", t)
            if lexer_conf.use_bytes:
                regexp = regexp.encode('utf-8')

            self.regexps[t.name] = lexer_conf.re_module.compile(regexp, lexer_conf.g_regex_flags)

    def match(self, term, text, index=0):
        return self.regexps[term.name].match(text, index)


def create_earley_parser__dynamic(lexer_conf: LexerConf, parser_conf: ParserConf, **kw):
    if lexer_conf.callbacks:
        raise GrammarError("Earley's dynamic lexer doesn't support lexer_callbacks.")

    earley_matcher = EarleyRegexpMatcher(lexer_conf)
    return xearley.Parser(lexer_conf, parser_conf, earley_matcher.match, **kw)

def _match_earley_basic(term, token):
    return term.name == token.type

def create_earley_parser__basic(lexer_conf: LexerConf, parser_conf: ParserConf, **kw):
    return earley.Parser(lexer_conf, parser_conf, _match_earley_basic, **kw)

def create_earley_parser(lexer_conf: LexerConf, parser_conf: ParserConf, options) -> earley.Parser:
    resolve_ambiguity = options.ambiguity == 'resolve'
    debug = options.debug if options else False
    tree_class = options.tree_class or Tree if options.ambiguity != 'forest' else None

    extra = {}
    if lexer_conf.lexer_type == 'dynamic':
        f = create_earley_parser__dynamic
    elif lexer_conf.lexer_type == 'dynamic_complete':
        extra['complete_lex'] = True
        f = create_earley_parser__dynamic
    else:
        f = create_earley_parser__basic

    return f(lexer_conf, parser_conf, resolve_ambiguity=resolve_ambiguity,
             debug=debug, tree_class=tree_class, ordered_sets=options.ordered_sets, **extra)



class CYK_FrontEnd:
    def __init__(self, lexer_conf, parser_conf, options=None):
        self.parser = cyk.Parser(parser_conf.rules)

        self.callbacks = parser_conf.callbacks

    def parse(self, lexer_thread, start):
        tokens = list(lexer_thread.lex(None))
        tree = self.parser.parse(tokens, start)
        return self._transform(tree)

    def _transform(self, tree):
        subtrees = list(tree.iter_subtrees())
        for subtree in subtrees:
            subtree.children = [self._apply_callback(c) if isinstance(c, Tree) else c for c in subtree.children]

        return self._apply_callback(tree)

    def _apply_callback(self, tree):
        return self.callbacks[tree.rule](tree.children)


_parser_creators['earley'] = create_earley_parser
_parser_creators['cyk'] = CYK_FrontEnd


def _construct_parsing_frontend(
        parser_type: _ParserArgType,
        lexer_type: _LexerArgType,
        lexer_conf,
        parser_conf,
        options
):
    assert isinstance(lexer_conf, LexerConf)
    assert isinstance(parser_conf, ParserConf)
    parser_conf.parser_type = parser_type
    lexer_conf.lexer_type = lexer_type
    if parser_type == 'earley' and lexer_type == 'template':
        return TemplateEarleyFrontend(lexer_conf, parser_conf, options)
    return ParsingFrontend(lexer_conf, parser_conf, options)



from __future__ import absolute_import

import unittest
from functools import partial, reduce, partialmethod
from operator import add, mul
from unittest import TestCase
import copy
import pickle
import functools

from t_lark.tree import Tree
from t_lark.lexer import Token
from t_lark.visitors import Visitor, Visitor_Recursive, Transformer, Interpreter, visit_children_decor, v_args, Discard, Transformer_InPlace, \
    Transformer_InPlaceRecursive, Transformer_NonRecursive, merge_transformers


class TestTrees(TestCase):
    def setUp(self):
        self.tree1 = Tree('a', [Tree(x, y) for x, y in zip('bcd', 'xyz')])
        self.tree2 = Tree('a', [
            Tree('b', [Token('T', 'x')]),
            Tree('c', [Token('T', 'y')]),
            Tree('d', [Tree('z', [Token('T', 'zz'), Tree('zzz', 'zzz')])]),
        ])

    def test_eq(self):
        assert self.tree1 == self.tree1
        assert self.tree1 != 0

    def test_copy(self):
        assert self.tree1 == copy.copy(self.tree1)

    def test_deepcopy(self):
        assert self.tree1 == copy.deepcopy(self.tree1)

    def test_pickle(self):
        s = copy.deepcopy(self.tree1)
        data = pickle.dumps(s, protocol=pickle.HIGHEST_PROTOCOL)
        assert pickle.loads(data) == s

    def test_repr_runnable(self):
        assert self.tree1 == eval(repr(self.tree1))

    def test_iter_subtrees(self):
        expected = [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z'),
                    Tree('a', [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')])]
        nodes = list(self.tree1.iter_subtrees())
        self.assertEqual(nodes, expected)

    def test_iter_subtrees_topdown(self):
        expected = [Tree('a', [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')]),
                    Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')]
        nodes = list(self.tree1.iter_subtrees_topdown())
        self.assertEqual(nodes, expected)

    def test_find_token(self):
        expected = [Token('T', 'x'), Token('T', 'y'), Token('T', 'zz')]
        tokens = list(self.tree2.find_token('T'))
        self.assertEqual(tokens, expected)

    def test_visitor(self):
        class Visitor1(Visitor):
            def __init__(self):
                self.nodes=[]

            def __default__(self,tree):
                self.nodes.append(tree)
        class Visitor1_Recursive(Visitor_Recursive):
            def __init__(self):
                self.nodes=[]

            def __default__(self,tree):
                self.nodes.append(tree)

        visitor1=Visitor1()
        visitor1_recursive=Visitor1_Recursive()

        expected_top_down = [Tree('a', [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')]),
                    Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')]
        expected_botton_up= [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z'),
                    Tree('a', [Tree('b', 'x'), Tree('c', 'y'), Tree('d', 'z')])]

        visitor1.visit(self.tree1)
        self.assertEqual(visitor1.nodes,expected_botton_up)

        visitor1_recursive.visit(self.tree1)
        self.assertEqual(visitor1_recursive.nodes,expected_botton_up)

        visitor1.nodes=[]
        visitor1_recursive.nodes=[]

        visitor1.visit_topdown(self.tree1)
        self.assertEqual(visitor1.nodes,expected_top_down)

        visitor1_recursive.visit_topdown(self.tree1)
        self.assertEqual(visitor1_recursive.nodes,expected_top_down)

    def test_interp(self):
        t = Tree('a', [Tree('b', []), Tree('c', []), 'd'])

        class Interp1(Interpreter):
            def a(self, tree):
                return self.visit_children(tree) + ['e']

            def b(self, tree):
                return 'B'

            def c(self, tree):
                return 'C'

        self.assertEqual(Interp1().visit(t), list('BCde'))

        class Interp2(Interpreter):
            @visit_children_decor
            def a(self, values):
                return values + ['e']

            def b(self, tree):
                return 'B'

            def c(self, tree):
                return 'C'

        self.assertEqual(Interp2().visit(t), list('BCde'))

        class Interp3(Interpreter):
            def b(self, tree):
                return 'B'

            def c(self, tree):
                return 'C'

        self.assertEqual(Interp3().visit(t), list('BCd'))

    def test_transformer(self):
        t = Tree('add', [Tree('sub', [Tree('i', ['3']), Tree('f', ['1.1'])]), Tree('i', ['1'])])

        class T(Transformer):
            i = v_args(inline=True)(int)
            f = v_args(inline=True)(float)

            sub = lambda self, values: values[0] - values[1]

            def add(self, values):
                return sum(values)

        res = T().transform(t)
        self.assertEqual(res, 2.9)

        @v_args(inline=True)
        class T(Transformer):
            i = int
            f = float
            sub = lambda self, a, b: a-b

            def add(self, a, b):
                return a + b


        res = T().transform(t)
        self.assertEqual(res, 2.9)


        @v_args(inline=True)
        class T(Transformer):
            i = int
            f = float
            from operator import sub, add

        res = T().transform(t)
        self.assertEqual(res, 2.9)

    def test_vargs(self):
        @v_args()
        class MyTransformer(Transformer):
            @staticmethod
            def integer(args):
                return 1 # some code here

            @classmethod
            def integer2(cls, args):
                return 2 # some code here

            hello = staticmethod(lambda args: 'hello')

        x = MyTransformer().transform( Tree('integer', [2]))
        self.assertEqual(x, 1)
        x = MyTransformer().transform( Tree('integer2', [2]))
        self.assertEqual(x, 2)
        x = MyTransformer().transform( Tree('hello', [2]))
        self.assertEqual(x, 'hello')

    def test_smart_decorator(self):
        class OtherClass:
            @staticmethod
            def ab_staticmethod(a, b):
                return (a, b)

            @classmethod
            def ab_classmethod(cls, a, b):
                assert cls is OtherClass, cls
                return (a, b)

            def ab_method(self, a, b):
                assert isinstance(self, OtherClass), self
                return (a, b)

        @v_args(meta=True)
        class OtherTransformer(Transformer):
            @staticmethod
            def ab_staticmethod(meta, children):
                return tuple(children)

            @classmethod
            def ab_classmethod(cls, meta, children):
                assert cls is OtherTransformer, cls
                return tuple(children)

            def ab_method(self, meta, children):
                assert isinstance(self, OtherTransformer), self
                return tuple(children)

        class CustomCallable:
            def __call__(self, *args, **kwargs):
                assert isinstance(self, CustomCallable)
                return args

        oc_instance = OtherClass()
        ot_instance = OtherTransformer()

        def ab_for_partialmethod(self, a, b):
            assert isinstance(self, TestCls)
            return a, b

        @v_args(inline=True)
        class TestCls(Transformer):
            @staticmethod
            def ab_staticmethod(a, b):
                return (a, b)

            @classmethod
            def ab_classmethod(cls, a, b):
                assert cls is TestCls
                return (a, b)

            def ab_method(self, a, b):
                assert isinstance(self, TestCls)
                return (a, b)

            oc_class_ab_staticmethod = oc_instance.ab_staticmethod
            oc_class_ab_classmethod = oc_instance.ab_classmethod

            oc_ab_staticmethod = oc_instance.ab_staticmethod
            oc_ab_classmethod = oc_instance.ab_classmethod
            oc_ab_method = oc_instance.ab_method

            ot_class_ab_staticmethod = ot_instance.ab_staticmethod
            ot_class_ab_classmethod = ot_instance.ab_classmethod

            ot_ab_staticmethod = ot_instance.ab_staticmethod
            ot_ab_classmethod = ot_instance.ab_classmethod
            ot_ab_method = ot_instance.ab_method

            ab_partialmethod = partialmethod(ab_for_partialmethod, 1)
            set_union = set(["a"]).union
            static_add = staticmethod(add)
            partial_reduce_mul = staticmethod(partial(reduce, mul))

            custom_callable = CustomCallable()

        test_instance = TestCls()
        expected = {
            "ab_classmethod": ([1, 2], (1, 2)),
            "ab_staticmethod": ([1, 2], (1, 2)),
            "ab_method": ([1, 2], (1, 2)),
            "oc_ab_classmethod": ([1, 2], (1, 2)),
            "oc_class_ab_classmethod": ([1, 2], (1, 2)),

            # AFAIK, these two cases are impossible to deal with. `oc_instance.ab_staticmethod` returns an actual
            # function object that is impossible to distinguish from a normally defined method.
            # (i.e. `staticmethod(f).__get__(?, ?) is f` is True)
            # "oc_ab_staticmethod": ([1, 2], (1, 2)),
            # "oc_class_ab_staticmethod": ([1, 2], (1, 2)),

            "oc_ab_method": ([1, 2], (1, 2)),
            "ot_ab_classmethod": ([1, 2], (1, 2)),
            "ot_class_ab_classmethod": ([1, 2], (1, 2)),

            # Same as above
            # "ot_ab_staticmethod": ([1, 2], (1, 2)),
            # "ot_class_ab_staticmethod": ([1, 2], (1, 2)),

            "ot_ab_method": ([1, 2], (1, 2)),
            "ab_partialmethod": ([2], (1, 2)),
            "custom_callable": ([1, 2], (1, 2)),
            "set_union": ([["b"], ["c"]], {"a", "b", "c"}),
            "static_add": ([1, 2], 3),
            "partial_reduce_mul": ([[1, 2]], 2),
        }
        non_static = {"ab_method", "ab_partialmethod"}
        for method_name, (children, expected_result) in expected.items():
            not_inline = "ot" in method_name
            result = test_instance.transform(Tree(method_name, children))
            self.assertEqual(result, expected_result)

            if not_inline:
                result = getattr(test_instance, method_name)(None, children)
            else:
                result = getattr(test_instance, method_name)(*children)
            self.assertEqual(result, expected_result)

            if method_name not in non_static:
                if not_inline:
                    result = getattr(TestCls, method_name)(None, children)
                else:
                    result = getattr(TestCls, method_name)(*children)
                self.assertEqual(result, expected_result)

    def test_vargs_set_name(self):
        # Test with cached_property if available. That actually uses __set_name__
        prop = getattr(functools, "cached_property", property)

        class T(Transformer):
            @v_args(inline=True)
            @prop  # Not sure why you would ever want to use a property here, but we support it
            def test(self):
                return lambda a, b: (self, a, b)

        t = T()
        self.assertEqual(t.transform(Tree("test", [1, 2])), (t, 1, 2))

    def test_inline_static(self):
        @v_args(inline=True)
        class T(Transformer):
            @staticmethod
            def test(a, b):
                return a + b
        x = T().transform(Tree('test', ['a', 'b']))
        self.assertEqual(x, 'ab')

    def test_vargs_override(self):
        t = Tree('add', [Tree('sub', [Tree('i', ['3']), Tree('f', ['1.1'])]), Tree('i', ['1'])])

        @v_args(inline=True)
        class T(Transformer):
            i = int
            f = float
            sub = lambda self, a, b: a-b

            not_a_method = {'other': 'stuff'}

            @v_args(inline=False)
            def add(self, values):
                return sum(values)

        res = T().transform(t)
        self.assertEqual(res, 2.9)

    def test_partial(self):

        tree = Tree("start", [Tree("a", ["test1"]), Tree("b", ["test2"])])

        def test(prefix, s, postfix):
            return prefix + s.upper() + postfix

        @v_args(inline=True)
        class T(Transformer):
            a = staticmethod(functools.partial(test, "@", postfix="!"))
            b = staticmethod(functools.partial(lambda s: s + "!"))

        res = T().transform(tree)
        assert res.children == ["@TEST1!", "test2!"]

    def test_discard(self):
        class MyTransformer(Transformer):
            def a(self, args):
                return 1 # some code here

            def b(cls, args):
                return Discard

        t = Tree('root', [
            Tree('b', []),
            Tree('a', []),
            Tree('b', []),
            Tree('c', []),
            Tree('b', []),
        ])
        t2 = Tree('root', [1, Tree('c', [])])

        x = MyTransformer().transform( t )
        self.assertEqual(x, t2)

    def test_transformer_variants(self):
        tree = Tree('start', [
            Tree('add', [Token('N', '1'), Token('N', '2'), Token('IGNORE_TOKEN', '4')]),
            Tree('add', [Token('N', '3'), Token('N', '4')]),
            Tree('ignore_tree', [Token('DO', 'NOT PANIC')]),
            ])
        for base in (Transformer, Transformer_InPlace, Transformer_NonRecursive, Transformer_InPlaceRecursive):
            class T(base):
                def add(self, children):
                    return sum(children)

                def N(self, token):
                    return int(token)

                def ignore_tree(self, children):
                    return Discard

                def IGNORE_TOKEN(self, token):
                    return Discard

            copied = copy.deepcopy(tree)
            result = T().transform(copied)
            self.assertEqual(result, Tree('start', [3, 7]))

    def test_merge_transformers(self):
        tree = Tree('start', [
            Tree('main', [
                Token("A", '1'), Token("B", '2')
            ]),
            Tree("module__main", [
                Token("A", "2"), Token("B", "3")
            ])
        ])

        class T1(Transformer):
            A = int
            B = int
            main = sum
            start = list
            def module__main(self, children):
                return sum(children)

        class T2(Transformer):
            A = int
            B = int
            main = sum
            start = list

        class T3(Transformer):
            def main(self, children):
                return sum(children)

        class T4(Transformer):
            main = sum


        t1_res = T1().transform(tree)
        composed_res = merge_transformers(T2(), module=T3()).transform(tree)
        self.assertEqual(t1_res, composed_res)

        composed_res2 = merge_transformers(T2(), module=T4()).transform(tree)
        self.assertEqual(t1_res, composed_res2)

        with self.assertRaises(AttributeError):
            merge_transformers(T1(), module=T3())

    def test_transform_token(self):
        class MyTransformer(Transformer):
            def INT(self, value):
                return int(value)

        t = Token('INT', '123')
        assert MyTransformer().transform(t) == 123

        class MyTransformer(Transformer):
            def INT(self, value):
                return Discard

        assert MyTransformer().transform(t) is None


if __name__ == '__main__':
    unittest.main()
